package com.sinusgear
{
	import flash.events.Event;
	
	import mx.collections.ArrayCollection;
	import mx.rpc.events.FaultEvent;
	import mx.rpc.events.ResultEvent;
	import mx.rpc.http.HTTPService;
	
	import org.flixel.*;
	
	public class GameState extends FlxState
	{
		[Embed(source="../../../assets/ice-tiles.png")] 
		protected var TechTilesImage:Class;
		
		[Embed(source="../../../assets/coins.png")] 
		protected var CoinsImage:Class;
		
		[Embed(source="../../../assets/pickup.mp3")]
		protected var PickupSFX:Class;
		
		[Embed(source="../../../assets/wind-chimes.mp3")]
		protected var LandSFX:Class;
		
		[Embed(source="../../../assets/noriko-outro.mp3")]
		protected var OutroSFX:Class;
		
		[Embed(source="../../../assets/star-tiles.png")]
		protected var StarTiles:Class;
		
		protected var levelBlocks:Array = new Array();
		protected var quickCollisionMap:Array = new Array();
		protected var player:Player = null;
		protected var mapLayers:Array = null;

		/*[Embed(source="../assets/snowtree_32.png")]
		protected var TitleImage:Class;*/
		
		protected static const ENEMY_COUNT:uint = 10;
		protected var enemies:Array = new Array();
		
		protected static const RANDOM_BLOCK_MIN_DIMENSION:uint = 2;
		protected static const RANDOM_BLOCK_MAX_DIMENSION:uint = 6;
		protected static const LEVEL_DIMENSIONS:uint = 640;
		protected static const BLOCK_DIMENSIONS:uint = 32;
		protected static const NUMBER_RANDOM_BLOCKS:uint = 150;
		protected static const MAXIMUM_RANDOM_PLACEMENT_ATTEMPTS:uint = 5;
		
		protected var coins:Array = new Array();
		protected var coinCount:uint = 0;
		protected var coinCountText:FlxText = null;
		
		protected static const COINS_COUNT:uint = 10;
		
		protected var availableCoins:uint;
		
		protected var actualLevel:uint = 0;

		
		public function addBlock(x:uint, y:uint, w:uint, h:uint):void
		{
			var block:FlxBlock = new FlxBlock(x,y,w,h);
			block.loadGraphic(TechTilesImage);
			levelBlocks.push(this.add(block));
		}
		
		public function addBorders():void
		{
			this.addBlock(
				0,
				LEVEL_DIMENSIONS-BLOCK_DIMENSIONS,
				LEVEL_DIMENSIONS,
				BLOCK_DIMENSIONS
				);
			this.addBlock(
				0,
				0,
				LEVEL_DIMENSIONS,
				BLOCK_DIMENSIONS);
			this.addBlock(
				0,
				BLOCK_DIMENSIONS,
				BLOCK_DIMENSIONS,
				640-BLOCK_DIMENSIONS*2);
			this.addBlock(
				LEVEL_DIMENSIONS-BLOCK_DIMENSIONS,
				BLOCK_DIMENSIONS,
				BLOCK_DIMENSIONS,
				640-BLOCK_DIMENSIONS*2);

		}
		
		protected function addRandomBlocks():void
		{
			var blockWidth:uint = 0;
			var blockHeight:uint = 0;
			var blockX:uint = 0;
			var blockY:uint = 0;
			var collisionWithExistingBlocks:Boolean = false;
			var newRandomBlock:FlxBlock = null;
			var loopCount:uint = 0;
			
			do
			{
				collisionWithExistingBlocks = false;
				
				blockWidth = RANDOM_BLOCK_MIN_DIMENSION + int(FlxG.random() * 
					(RANDOM_BLOCK_MAX_DIMENSION - RANDOM_BLOCK_MIN_DIMENSION));
				blockHeight = RANDOM_BLOCK_MIN_DIMENSION + int(FlxG.random() * 
					(RANDOM_BLOCK_MAX_DIMENSION - RANDOM_BLOCK_MIN_DIMENSION));
				blockX = int(FlxG.random() * LEVEL_DIMENSIONS / BLOCK_DIMENSIONS - RANDOM_BLOCK_MAX_DIMENSION) * 
					BLOCK_DIMENSIONS;
				blockY = int(FlxG.random() * LEVEL_DIMENSIONS / BLOCK_DIMENSIONS - RANDOM_BLOCK_MAX_DIMENSION) * 
					BLOCK_DIMENSIONS;
				
				newRandomBlock = new FlxBlock(
					blockX, 
					blockY, 
					blockWidth * BLOCK_DIMENSIONS, 
					blockHeight * BLOCK_DIMENSIONS);
				
				for each (var existingBlock:FlxBlock in levelBlocks)
				{
					collisionWithExistingBlocks = existingBlock.overlaps(newRandomBlock);
					if (collisionWithExistingBlocks) break;
				}
				
				++loopCount;
				
			} while (collisionWithExistingBlocks && loopCount < MAXIMUM_RANDOM_PLACEMENT_ATTEMPTS);
			newRandomBlock.loadGraphic(TechTilesImage);
			levelBlocks.push(this.add(newRandomBlock));
		}
		
		protected function isCollisionWithBlocks(blockX:uint, blockY:uint):Boolean
		{
			// Do not allow to go away from boundaries.
			if ((blockX >= 640) || (blockY >= 640))
			{
				return true;
			}
			
			if (this.quickCollisionMap[blockX/32 + (blockY/32)*20] == 1)
			{
				return true;
			}
			return false;
		}
		
		protected function addCoins():void
		{
			var i:uint;
			var blockX:uint = 0;
			var blockY:uint = 0;
			
			for (i=0; i<COINS_COUNT; i++)
			{
				do 
				{
				blockX = int(FlxG.random() * LEVEL_DIMENSIONS / BLOCK_DIMENSIONS - RANDOM_BLOCK_MAX_DIMENSION) * 
					BLOCK_DIMENSIONS;
				blockY = int(FlxG.random() * LEVEL_DIMENSIONS / BLOCK_DIMENSIONS - RANDOM_BLOCK_MAX_DIMENSION) * 
					BLOCK_DIMENSIONS;
				} while(isCollisionWithBlocks(blockX, blockY));
					
				var coin:Coin = new Coin(blockX, blockY);
				coin.loadGraphic(CoinsImage,true,false,32,32);
				coins.push(this.add(coin));
				this.availableCoins++;
			}
		}

		
		public function showLayer(index:int):void
		{
			var i:int;
			var layer:Array = this.mapLayers[index];
			// Map is linear so we have to keep track
			// of coordinates
			var x:uint;
			var y:uint;
			
			for(i=0; i < layer.length; i++)
			{
				if (layer[i] == '1')
				{
					this.quickCollisionMap[i] = 1;
					this.addBlock(x*32,y*32,32,32);
				}
				
				x++;
				
				// Move to next line
				if (x>=20)
				{
					x = 0;
					y++;
				}
				
			}
		}
		
		public function loadLevelFaultHandler(event:Event):void
		{
			trace("Unable to load level definition.");
		}
		
		public function loadLevelCompleteHandler(event:ResultEvent = null):void
		{
			var layers:ArrayCollection = event.result.TilemapFile.tilemap.layer;
			
			var i:int;
			
			this.mapLayers = new Array();
			
			// Parse data block from maps.
			for (i=0; i<layers.length; i++)
			{
				var data:String = layers[i].data;
				var myPattern:RegExp = /[ |\n|\r|\t]/g;
				data = data.replace(myPattern,'');
				
				var dataMap:Array = data.split(',');
				this.mapLayers.push(dataMap);
			}
			
			// Clean quick collision map
			for (i=0; i < 20*20; i++)
			{
				this.quickCollisionMap.push(0);
			}
			
			//this.addBorders();
			this.showLayer(0);
			
			coinCountText = new FlxText(10, 10, FlxG.width, "* 0");
			coinCountText.size = 14;
			coinCountText.color = 0xffd71d;
			coinCountText.scrollFactor.x = coinCountText.scrollFactor.y = 0;
			this.add(coinCountText);
			
			this.addCoins();
			player = new Player();
			this.add(player);
			FlxG.follow(player,2.5);
			FlxG.followAdjust(0.5,0.0);
			FlxG.followBounds(0,0,640,640);
			/*this.add(new FlxSprite(0,0,TitleImage));
			this.add(new FlxSprite(1,0,TitleImage));
			this.add(new FlxSprite(2,0,TitleImage));
			this.add(new FlxSprite(10,0,TitleImage));*/

		}
		
		public function loadLevel(path:String):void
		{
			
			var loader:HTTPService = new HTTPService();
			loader.url = path;
			loader.addEventListener(FaultEvent.FAULT, loadLevelFaultHandler);
			loader.addEventListener(ResultEvent.RESULT, loadLevelCompleteHandler);
			
			loader.send();
			
		}
		
		public function displayBackground():void
		{
			var block:FlxBlock = new FlxBlock(0,0,640,640);
			block.loadGraphic(StarTiles);
			this.add(block);
		}

		
		public function GameState()
		{
			this.displayBackground();
			this.loadLevel('global.xml');
			
			//for (var j:uint = 0; j < NUMBER_RANDOM_BLOCKS; ++j)
			//	addRandomBlocks();

			//this.addBlock(0,640-24,640,32);
			
		}

		protected function coinPickup(Collide1:FlxCore, Collide2:FlxCore):void
		{
			this.availableCoins--;
			++coinCount;
			
			
			coinCountText.text = "* " + coinCount.toString();
			Collide1.exists = false;
			FlxG.play(PickupSFX);   

			// All coins collected
			if (this.availableCoins == 0)
			{
				// We still have something to play ;)
				if (this.actualLevel<3)
				{
					this.actualLevel++;
					this.showLayer(actualLevel);
					this.addCoins();
					FlxG.play(LandSFX);   
				} else {
					FlxG.switchState(OutroState);
					FlxG.play(OutroSFX);
				}
				


			}

		}

		
		public override function update():void
		{
			super.update();
			FlxG.collideArray(levelBlocks, player);
			FlxG.collideArrays(levelBlocks, coins);
			FlxG.overlapArray(coins, player, this.coinPickup);

		}
		
	}
}
